// CommQueue.cpp : implementation file
//
// Copyright(C) 2009 ubisys technologies GmbH, Duesseldorf, Germany.
// All rights reserved.
//
// www.ubisys.de
// support@ubisys.de
//
// THIS SOURCE CODE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//
// IN NO EVENT SHALL UBISYS TECHNOLOGIES BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOURCE CODE OR THE USE OR
// OTHER DEALINGS IN THE SOURCE CODE.
//
// Part of the ubisys(R) 13.56 MHz RFID USB stick demo application
//
// Provides CCommQueue, a CCommunicationsQueue derived class that provides
// multi-threaded data exchange via serial ports or USB CDC/ACM devices,
// such as the 13.56 MHz RFID stick 

#include <pthread.h>
#include <fcntl.h>
#include <termios.h>
#include <sys/ioctl.h>
#include <cstdlib>
#include <csignal>
#include <cerrno>
#include <string>
#include "CompactFramework.h"
#include "Packet.h"
#include "CommunicationsQueue.h"
#include "CommQueue.h"

using namespace std;


/////////////////////////////////////////////////////////////////////////////
// CCommQueue

CCommQueue::CCommQueue() : CCommunicationsQueue(ReadThread, WriteThread), 
m_hFile(-1)
{
}


CCommQueue::~CCommQueue()
{
	// First of all, terminate threads and wait until they have terminated
	TerminateThreads();

	// Close the file handle
	Close();
}


bool CCommQueue::Open(const char *const pszPort, const int nDesiredAccess)
{
	m_strDevice = pszPort;
	
	// Open comm port for overlapped read/write access
	m_hFile = open(pszPort, nDesiredAccess);

#ifdef _DEBUG
	if (m_hFile == -1)
		TRACE2("CCommQueue::Open(%s) failed: %u\n", pszPort, errno);
#endif

	return (m_hFile != -1);
}


bool CCommQueue::Setup(const speed_t nBaudRate, 
	const tcflag_t cflagAdd, const tcflag_t cflagRemove,
	const tcflag_t iflagAdd, const tcflag_t iflagRemove,
	const tcflag_t oflagAdd, const tcflag_t oflagRemove)
{
	termios options;
		
	// Get current serial port settings
	tcgetattr(m_hFile, &options);

	// Set data rate to 1.5Mbps
	cfsetispeed(&options, nBaudRate);
	cfsetospeed(&options, nBaudRate);

	options.c_cflag &= ~cflagRemove;
	options.c_cflag |= cflagAdd;
	options.c_iflag &= ~iflagRemove;
	options.c_iflag |= iflagAdd;
	options.c_oflag &= ~oflagRemove;
	options.c_oflag |= oflagAdd;
	cfmakeraw(&options);

	// Set new serial port settings
	tcsetattr(m_hFile, TCSANOW, &options);
	
	return true;
}


bool CCommQueue::Close()
{
	return close(m_hFile) ? false : true;
}


void *CCommQueue::ReadThread(void *pArgument)
{
	CCommQueue *pThis = static_cast<CCommQueue *>(pArgument);

	// Setup terminate handler for this thread
	SetupTerminateHandler();
	
	// Thread has started. Allow Initialize() to continue...
	pThis->m_bReadThreadStarted = true;

	// Call the actual implementation
	return pThis->ReadThread();
}


void *CCommQueue::WriteThread(void *pArgument)
{
	CCommQueue *pThis = static_cast<CCommQueue *>(pArgument);
	
	// Setup terminate handler for this thread
	SetupTerminateHandler();	
	
	// Thread has started. Allow Initialize() to continue...
	pThis->m_bWriteThreadStarted = true;
	
	// Call the actual implementation
	return pThis->WriteThread();
}


void *CCommQueue::ReadThread()
{
	fd_set r;
	
	FD_ZERO(&r);
	FD_SET(m_hFile, &r);
	
	TRACE0("CCommQueue RX thread started.\r\n");
	
	while (!m_bTerminate)
	{
		if ((select(m_hFile + 1, &r, 0, 0, 0) == 1) && m_bReadEnable)
		{
			// Determine number of bytes available in receive buffer
			size_t cbExpected = 0;
			
			if (ioctl(m_hFile, FIONREAD, &cbExpected) < 0)
				cbExpected = 0;
		
			CPacket data(cbExpected);
	
			const size_t cbActual = read(m_hFile, data.GetDataAt(data.GetSize()),
				data.GetCapacity() - data.GetSize());
			
			if (cbActual > 0)
				data.SetSize(data.GetSize() + cbActual);

			// Invoke transmit notification
			pthread_mutex_lock(&m_csr);
			
			ASSERT(data.GetSizeEx() == cbExpected);
			AfterReceive(data);
	
			// TODO: we could limit the size of the input queue
			if (m_receive.IsEmpty())
				m_receive = CPacket(data.GetData(), data.GetSize());
			else
				m_receive.Append(data);
				
			// If there is data available, inform waiting threads
			pthread_cond_signal(&m_conditionReceive);
			
			pthread_mutex_unlock(&m_csr);
		}	
	}

	TRACE0("CCommQueue RX thread terminating...\r\n");
		
	return 0;
}


void *CCommQueue::WriteThread()
{
	TRACE0("CCommQueue TX thread started.\r\n");
	
	while (!m_bTerminate)
	{
		pthread_mutex_lock(&m_csw);
	
		// Wait for data to write...
		pthread_cond_wait(&m_conditionTransmit, &m_csw);
		
		if (!m_bTerminate && m_bWriteEnable && m_transmit.GetSizeEx())
		{
			size_t nTransferred;
			
			do
			{
				nTransferred = write(m_hFile, m_transmit.GetData(), 
					m_transmit.GetSize());
					
				if (nTransferred > 0)
				{
					// Invoke transmit notification
					AfterTransmit(CPacket(m_transmit.GetData(), nTransferred));
					m_transmit.StripHeader(nTransferred);
				}
			}
			while (nTransferred > 0);
			
			// Clear TX output packet, if empty
			if (!m_transmit.GetSizeEx())
				m_transmit.Attach();
		}
		
		pthread_mutex_unlock(&m_csw);
	}

	TRACE0("CCommQueue TX thread terminating...\r\n");

	return 0;
}


void CCommQueue::AfterTransmit(const CPacket &packet)
{
	/*
#ifdef _DEBUG
	packet.Dump("TX");
#endif
	*/
}


void CCommQueue::AfterReceive(const CPacket &packet)
{
	/*
#ifdef _DEBUG
	packet.Dump("RX");
#endif
	*/
}


bool CCommQueue::PurgeReceptionBuffer()
{
	return tcflush(m_hFile, TCIFLUSH) ? false : true;
}


bool CCommQueue::PurgeTransmissionBuffer()
{
	return tcflush(m_hFile, TCOFLUSH) ? false : true;
}

/////////////////////////////////////////////////////////////////////////////
// Class GUID for plug & play serial ports (serenum driver)

/*
#include <initguid.h>

#ifndef GUID_DEVINTERFACE_SERENUM_BUS_ENUMERATOR
// 4D36E978-E325-11CE-BFC1-08002BE10318
DEFINE_GUID(GUID_DEVINTERFACE_SERENUM_BUS_ENUMERATOR, 0x4d36e978, 0xe325, \
	0x11ce, 0xbf, 0xc1, 0x08, 0x00, 0x2b, 0xe1, 0x03, 0x18);
#endif

bool CCommQueue::GetStringProperty(HDEVINFO hDeviceInfo, 
	PSP_DEVINFO_DATA pDeviceInfoData, DWORD dwProperty, CString &str)
{
	DWORD dwSize, dwType;
	bool bResult;

	// Determine required size
	bResult = !SetupDiGetDeviceRegistryProperty(hDeviceInfo, pDeviceInfoData,
		dwProperty, &dwType, 0, 0, &dwSize) && 
		(GetLastError() == ERROR_INSUFFICIENT_BUFFER);

	bResult = bResult && (dwType == REG_SZ);

	// Allocate storage for desired property, lock buffer, retrieve property
	bResult = bResult && SetupDiGetDeviceRegistryProperty(hDeviceInfo, 
		pDeviceInfoData, dwProperty, NULL, 
		reinterpret_cast<PBYTE>(str.GetBufferSetLength(dwSize)), dwSize, NULL);

	// Unlock buffer, update string size
	str.ReleaseBuffer();

	return bResult;
}


bool CCommQueue::GetStringProperty(HDEVINFO hDeviceInfo, 
	PSP_DEVINFO_DATA pDeviceInfoData, DWORD dwProperty, CStringList &list)
{
	DWORD dwSize, dwType;
	bool bResult;
	CString str;

	// Determine required size
	bResult = !SetupDiGetDeviceRegistryProperty(hDeviceInfo, pDeviceInfoData,
		dwProperty, &dwType, 0, 0, &dwSize) && 
		(GetLastError() == ERROR_INSUFFICIENT_BUFFER);

	bResult = bResult && (dwType == REG_MULTI_SZ);

	// Convert size into number of characters (i.e. length)
	const UINT nLength = dwSize / sizeof(TCHAR);

	// Allocate storage for desired property, lock buffer, retrieve property
	bResult = bResult && SetupDiGetDeviceRegistryProperty(hDeviceInfo, 
		pDeviceInfoData, dwProperty, NULL, 
		reinterpret_cast<PBYTE>(str.GetBufferSetLength(nLength)), dwSize, NULL);

	// Unlock buffer, update string size
	str.ReleaseBufferSetLength(nLength);

	ENSURE((str.GetLength() >= 3) && (str[str.GetLength() - 2] == _T('\0')) 
		&& (str[str.GetLength() - 1] == _T('\0')));

	list.AddTail(CString());

	for (int i = 0; i < str.GetLength() - 2; i++)
	{
		if (str[i] != _T('\0'))
			list.GetTail() += str[i];
		else
			list.AddTail(CString());
	}

	return bResult;
}


bool CCommQueue::Enumerate(std::vector<CCommPortInfo> &ports, 
	const bool bIgnoreBusy)
{
	HDEVINFO hDevInfo = 
		SetupDiGetClassDevs(&GUID_DEVINTERFACE_SERENUM_BUS_ENUMERATOR, NULL, 
		NULL, DIGCF_PRESENT | DIGCF_DEVICEINTERFACE);
	
	// Clear result vector
	ports.clear();

	if (hDevInfo == INVALID_HANDLE_VALUE)
	{
		TRACE1("SetupDiGetClassDevs failed: %s\n", 
			GetSystemErrorMessage(GetLastError()));
		
		return false;
	}

	bool bResult = true;

	for (DWORD dwItem = 0; bResult; dwItem++)
	{
		SP_DEVICE_INTERFACE_DATA ifcData;

		ifcData.cbSize = sizeof(SP_DEVICE_INTERFACE_DATA);

		bResult = SetupDiEnumDeviceInterfaces(hDevInfo, NULL, 
			&GUID_DEVINTERFACE_SERENUM_BUS_ENUMERATOR, dwItem, &ifcData) ? 
			true : false;

		if (!bResult)
		{
			bResult = (GetLastError() == ERROR_NO_MORE_ITEMS);
			break;
		}

		// Got a device. Get the details.
		DWORD cbDetail;

		// Determine required detail size
		bResult = bResult && !SetupDiGetDeviceInterfaceDetail(hDevInfo, 
			&ifcData, 0, 0, &cbDetail, 0) && 
			(GetLastError() == ERROR_INSUFFICIENT_BUFFER);

		if (bResult)
		{
			SP_DEVINFO_DATA devinfo;
			SP_DEVICE_INTERFACE_DETAIL_DATA *pDetail = 
				reinterpret_cast<SP_DEVICE_INTERFACE_DETAIL_DATA *>(new char[cbDetail]);

			devinfo.cbSize = sizeof(SP_DEVINFO_DATA);
			pDetail->cbSize = sizeof(SP_DEVICE_INTERFACE_DETAIL_DATA);

			bResult = bResult && SetupDiGetDeviceInterfaceDetail(hDevInfo, &ifcData, 
				pDetail, cbDetail, NULL, &devinfo);

			if (bResult)
			{
				CCommPortInfo info;

				info.m_strDevicePath = pDetail->DevicePath;

				bResult = bResult && GetStringProperty(hDevInfo, &devinfo, 
					SPDRP_FRIENDLYNAME, info.m_strFriendly);

				bResult = bResult && GetStringProperty(hDevInfo, &devinfo, 
					SPDRP_DEVICEDESC, info.m_strDescription);

				bResult = bResult && GetStringProperty(hDevInfo, &devinfo, 
					SPDRP_HARDWAREID, info.m_listHardwareIDs);

				// If all required information has been obtained, add a new entry
				if (bResult) 
					ports.push_back(info);
			}

			delete[] reinterpret_cast<char *>(pDetail);
		}
	}

	SetupDiDestroyDeviceInfoList(hDevInfo);

	if (bIgnoreBusy)
		RejectBusyDevices(ports);

	return bResult;
}


bool CCommQueue::SelectUSBDevices(std::vector<CCommPortInfo> &ports, 
	const WORD wProductID, const WORD wVendorID)
{
	// Walk the array of enumerated devices and filter out those that
	// are no USB devices or whose VID/PID do not match the requested
	// values
	for (std::vector<CCommPortInfo>::iterator i = ports.begin(); i != ports.end(); )
	{
		WORD wThisVendorID, wThisProductID;
		CString strSerialNumber;

		// Unable to obtain USB device information? VID and PID match?
		const bool bReject = !GetUSBDeviceInformation(i->m_strDevicePath, 
			wThisVendorID, wThisProductID, strSerialNumber) || 
			(wVendorID && ((wThisProductID != wProductID) || 
				(wThisVendorID != wVendorID)));

		// If entry is to be rejected, 
		if (bReject)
			i = ports.erase(i);
		else
			i++;
	}

	return true;
}


void CCommQueue::RejectBusyDevices(std::vector<CCommPortInfo> &ports)
{
	// Remove those ports that cannot be opened for read and write
	for (std::vector<CCommPortInfo>::size_type i = 0; i < ports.size(); )
	{
		// Try to obtain a device handle
		HANDLE hDevice = CreateFile(ports[i].m_strDevicePath,
			GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, 
			FILE_FLAG_OVERLAPPED, NULL);

		// If handle is valid, close it first
		if (hDevice != INVALID_HANDLE_VALUE)
			CloseHandle(hDevice);

		// Device is available, when handle was valid
		ports[i].m_bAvailable = (hDevice != INVALID_HANDLE_VALUE);

		if (!ports[i].m_bAvailable)
			ports.erase(ports.begin() + i);
		else
			i++;
	}
}


bool CCommQueue::GetUSBDeviceInformation(const CString &strDevicePath, 
	WORD &wVendorID, WORD &wProductID, CString &strSerialNumber)
{
	CString strToken, strValue;
	LPTSTR pszEnd;
	int nPosition = 0;

	// Extract serial number from fully qualified device path. A typical 
	// device path looks like this:
	// "\\?\usb#vid_19a6&pid_6124#0123456789ab#{a3de00ea-eb56-479c-91b2-977c7b6a4503}"
	strToken = strDevicePath.Tokenize(_T("#"), nPosition);

	if (strToken != _T("\\\\?\\usb"))
		return false;

	// Vendor ID & Product ID
	strToken = strDevicePath.Tokenize(_T("#"), nPosition);

	if ((strToken.GetLength() != 17) || (strToken.Left(4) != _T("vid_")))
		return false;

	strValue = strToken.Mid(4, 4);
	wVendorID = static_cast<unsigned short>(_tcstoul(strValue, &pszEnd, 16));

	if (pszEnd != static_cast<LPCTSTR>(strValue) + strValue.GetLength())
		return false;

	if (strToken.Mid(8, 5) != _T("&pid_"))
		return false;

	strValue = strToken.Right(4);
	wProductID = static_cast<unsigned short>(_tcstoul(strValue, &pszEnd, 16));

	if (pszEnd != static_cast<LPCTSTR>(strValue) + strValue.GetLength())
		return false;

	// Serial Number
	strSerialNumber = strDevicePath.Tokenize(_T("#"), nPosition);

	return true;
}
*/
