// CommunicationsQueue.cpp : implementation file
//
// Copyright(C) 2009 ubisys technologies GmbH, Duesseldorf, Germany.
// All rights reserved.
//
// www.ubisys.de
// support@ubisys.de
//
// THIS SOURCE CODE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//
// IN NO EVENT SHALL UBISYS TECHNOLOGIES BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOURCE CODE OR THE USE OR
// OTHER DEALINGS IN THE SOURCE CODE.
//
// Part of the ubisys(R) 13.56 MHz RFID USB stick demo application
//
// Provides CCommunicationsQueue, an abstract base class for multi-threaded 
// data exchange via pipes or other serializing communication links

#include <pthread.h>
#include <fcntl.h>
#include <cstdlib>
#include <csignal>
#include <stdexcept>
#include "CompactFramework.h"
#include "Packet.h"
#include "CommunicationsQueue.h"

using namespace std;

/////////////////////////////////////////////////////////////////////////////
// CCommunicationsQueue

CCommunicationsQueue::CCommunicationsQueue(LPTHREAD_START_ROUTINE pfnReadThread, 
	LPTHREAD_START_ROUTINE pfnWriteThread) :
m_pClient(0), m_hReadThread(0), m_hWriteThread(0), 
m_bReadEnable(false), m_bWriteEnable(false),
m_pfnReadThread(pfnReadThread), m_pfnWriteThread(pfnWriteThread),
m_bReadThreadStarted(false), m_bWriteThreadStarted(false),
m_bTerminate(false)
{
	// Initialize critical section objects
	pthread_mutex_init(&m_csr, 0);
	pthread_mutex_init(&m_csw, 0);
	pthread_cond_init(&m_conditionReceive, 0);
	pthread_cond_init(&m_conditionTransmit, 0);
}


CCommunicationsQueue::~CCommunicationsQueue()
{
	// Terminate the worker threads
	TerminateThreads();

	// Delete critical sections
	pthread_mutex_destroy(&m_csr);
	pthread_mutex_destroy(&m_csw);
	pthread_cond_destroy(&m_conditionReceive);
	pthread_cond_destroy(&m_conditionTransmit);
}


bool CCommunicationsQueue::Initialize(const int nDesiredAccess)
{
	ASSERT(!m_bReadThreadStarted && !m_bWriteThreadStarted);
	m_bTerminate = false;
	
	if (nDesiredAccess & (O_RDWR | O_RDONLY))
	{
		// Create the reading thread
		if (pthread_create(&m_hReadThread, 0, m_pfnReadThread, this))
			return false;
			
		// Wait until the reading thread has started up...
		TRACE0("Initialize(): Anticipating RX startup...\r\n");
		
		while (!m_bReadThreadStarted)
			usleep(100);
	}

	if (nDesiredAccess & (O_RDWR | O_WRONLY))
	{
		// Create the writing thread
        if (pthread_create(&m_hWriteThread, 0, m_pfnWriteThread, this))
			return false;
			
		// Wait until the reading thread has started up...
		TRACE0("Initialize(): Anticipating TX startup...\r\n");
					
		// Wait until the writing thread has started up...
		while (!m_bWriteThreadStarted)
			usleep(100);
	}

	return true;
}


void CCommunicationsQueue::OnSignalTerminate(int nSignal, siginfo_t *pInfo, void *pContext)
{
	// TRACE1("caught signal #%u\r\n", nSignal);
}


void CCommunicationsQueue::SetupTerminateHandler()
{
	struct sigaction sa;
	
	// Setup signal handler for termination request
	sa.sa_sigaction = &OnSignalTerminate;
	sa.sa_flags = SA_SIGINFO;
	sigemptyset(&sa.sa_mask);
	sigaction(SIGTERM, &sa, 0);	
}


bool CCommunicationsQueue::TerminateThreads()
{
	// Any thread running?
	if (m_hReadThread || m_hWriteThread)
	{
		m_bTerminate = true;
		
		if (m_hReadThread)
			pthread_kill(m_hReadThread, SIGTERM);

		if (m_hWriteThread)
		{
			pthread_mutex_lock(&m_csw);
			pthread_cond_signal(&m_conditionTransmit);
			pthread_mutex_unlock(&m_csw);			
		}

		if (m_hReadThread)
		{
			pthread_join(m_hReadThread, 0);
			TRACE0("CCommunicationsQueue RX thread terminated.\r\n");
		}

		if (m_hWriteThread)
		{
			pthread_join(m_hWriteThread, 0);
			TRACE0("CCommunicationsQueue TX thread terminated.\r\n");
		}

		m_hReadThread = 0;
		m_hWriteThread = 0;
	}

	return true;
}


bool CCommunicationsQueue::EnableReception(const bool bEnable, const bool bClear)
{
	// Notice: POSIX threads do not support suspend/resume
	
	if (m_bReadEnable != bEnable)
	{
		m_bReadEnable = bEnable;

		if (bEnable)
		{
			if (bClear)
			{
				PurgeReceptionBuffer();
				
				if (!m_receive.IsEmpty())
					m_receive.Attach();
			}

			// ResumeThread(m_hReadThread);
		}
		else
		{
			// When suspending the thread, make sure that the
			// thread is not currently reading
			/*
			EnterCriticalSection(&m_csr);
			SuspendThread(m_hReadThread);
			LeaveCriticalSection(&m_csr);
			*/
		}

		return true;
	}

	return false;
}


bool CCommunicationsQueue::EnableTransmission(const bool bEnable, const bool bClear)
{
	// Notice: POSIX threads do not support suspend/resume
	
	if (m_bWriteEnable != bEnable)
	{
		m_bWriteEnable = bEnable;

		if (bEnable)
		{
			if (bClear)
			{
				if (!m_transmit.IsEmpty())
					m_transmit.Attach();

				PurgeTransmissionBuffer();
			}

			// ResumeThread(m_hWriteThread);
		}
		else
		{
			// When suspending the thread, make sure that the
			// thread is not currently writing
			/*
			EnterCriticalSection(&m_csw);
			SuspendThread(m_hWriteThread);
			LeaveCriticalSection(&m_csw);
			*/
		}

		return true;
	}

	return false;
}


int CCommunicationsQueue::Receive(CPacket &data)
{
	// Gain access to receiver queue
	pthread_mutex_lock(&m_csr);

	// Retrieve data from input queue
	data = m_receive;
	m_receive.Attach();
	
	// Release access to receiver queue
	pthread_mutex_unlock(&m_csr);

	if (m_pClient && data.GetSizeEx())
		m_pClient->AppendLog(data, false);

	return static_cast<int>(data.GetSizeEx());
}


int CCommunicationsQueue::Send(const CPacket &data)
{
	int nResult;

	// Must not try to send empty data
	ASSERT(data.GetSizeEx());

	// Not enabled for writing? Return zero
	if (!m_bWriteEnable)
		return 0;

	if (m_pClient)
		m_pClient->AppendLog(data, true);

	// Gain access to transmitter queue
	pthread_mutex_lock(&m_csw);

	// Transfer data from data vector provided by caller
	// to output queue
	if (m_transmit.IsEmpty())
		m_transmit = CPacket(data.GetData(), data.GetSize());
	else
		m_transmit.Append(data);

	nResult = static_cast<int>(m_transmit.GetSize());
	
	// Instruct worker thread to commence transmission
	pthread_cond_signal(&m_conditionTransmit);
	
	// Release access to transmitter queue
	pthread_mutex_unlock(&m_csw);

	return nResult;
}
