// Packet.cpp : implementation file
//
// Copyright(C) 2008 ubisys technologies GmbH, Duesseldorf, Germany.
// All rights reserved.
//
// www.ubisys.de
// support@ubisys.de
//
// Part of the Compact15.4(TM) class library
//
// A powerful packet container class that provides easy packet assembly and 
// disassembly in an efficient manner. This implementation allows the stack
// implementor to reserve space before and after the body. This allows adding
// headers and trailers to the actual payload as the packet traverses the 
// stack. If enough space is reserved, no addional memory moving or copying
// is required. A copy-on-write feature further improves overall performance.
// This implementation is scalable. The stack designer may enable/disable 
// features individually depending on the runtime environment (e.g. RFD vs. 
// FFD device, tiny active RFID tag vs. PC, etc.).
//
// Also provides a means for future heap compacting algorithms, as a second
// level of indirection is introduced.
//
// __CPACKET_ALLOW_GROW__ enables automatic packet growth beyond initial 
// capacity limits (within capacity limits, automatic packet growth is always
// possible). __CPACKET_ALLOW_MOVE__ must be defined, too.
//
// __CPACKET_ALLOW_MOVE__ allows moving packet contents within the overall 
// reserved space. This allows to shift between header and trailer capacity
// as required at the expence of a call to memmove().
//
// #define __CPACKET_ALLOW_COW__, to allow copy-on-write on packet level. 
// Does not require __CPACKET_ALLOW_GROW__

#include <cstdlib>
#include <cstring>
#include "CompactFramework.h"
#include "Packet.h"

/////////////////////////////////////////////////////////////////////////////
// When __CPACKET_ALLOW_GROW__ is defined, __CPACKET_ALLOW_MOVE__
// must be defined, too

#if (defined __CPACKET_ALLOW_GROW__ && !defined __CPACKET_ALLOW_MOVE__)
#error If packets are allowed to grow, data must also be allowed to move
#endif


/////////////////////////////////////////////////////////////////////////////
// ASSERT_EX() macro

#ifdef _DEBUG
#define ASSERT_EX(obj, e) \
	{ \
		if (!(e)) \
		{ \
			(obj).Dump(); \
			ASSERT(false); \
		} \
	}
#else
#define ASSERT_EX(obj, e)
#endif


/////////////////////////////////////////////////////////////////////////////
// CPacketData

CPacketData::CPacketData(const arg_type nBodyAndTrailerCapacity, 
	const arg_type nHeaderCapacity) :
m_pbData(0), m_nSize(0), m_nCapacity(0), m_nOffset(nHeaderCapacity), 
m_nReference(0)
{
	// If one of these ASSERTions bombs, you passed a value that cannot be
	// held in a variable of type size_type
	ASSERT(nBodyAndTrailerCapacity <= static_cast<size_type>(-1));
	ASSERT(nHeaderCapacity <= static_cast<size_type>(-1));
	
	// Allocate enough storage for header, body and footer
	Reserve(nHeaderCapacity + nBodyAndTrailerCapacity);
}


CPacketData::CPacketData(const CPacketData &data) : 
m_pbData(0), m_nSize(0), m_nCapacity(0), m_nOffset(data.m_nOffset), 
m_nReference(0)
{
	// Allocate storage
	Reserve(data.m_nCapacity);

	// Append data
	Append(data.GetData(), data.GetSize());
}


CPacketData::~CPacketData()
{
	ASSERT_VALIDHEAPBLOCK(this, sizeof(CPacketData));
	ASSERT_EX(*this, m_pbData);
	ASSERT_VALIDHEAPBLOCK(m_pbData, m_nCapacity);
	free(m_pbData);
}


CPacketData::size_type CPacketData::AddReference()
{
	// Increment reference counter by one and return its current (i.e. new) 
	// value
	return ++m_nReference;
}


CPacketData::size_type CPacketData::Release()
{
	ASSERT_VALIDHEAPBLOCK(this, sizeof(CPacketData));	
	ASSERT_EX(*this, m_nReference > 0);
	const size_type nResult = --m_nReference;

	// If reference count has reached zero, delete this object now
	if (!nResult)
		delete this;

	return nResult;
}


void CPacketData::SetSize(const arg_type nSize)
{
	// If this ASSERTions bombs, you passed a value that cannot be
	// held in a variable of type size_type
	ASSERT(nSize <= static_cast<size_type>(-1));

#ifdef __CPACKET_ALLOW_GROW__
	// Allows packets to grow beyond the capacity reserved during 
	// construction
	if (nSize > m_nCapacity)
		Reserve(nSize);
#else
	// If you end up here, increase the inital capacity limit (header) for 
	// this packet, or allow packets to grow (not recommended for tiny 
	// devices)
	ASSERT_EX(*this, nSize <= m_nCapacity);
#endif

#ifdef __CPACKET_ALLOW_MOVE__
	// Move existing contents to make room for header
	if (nSize > m_nCapacity - m_nOffset)
	{
		const size_type nOffset = m_nCapacity - nSize;

		memmove(&m_pbData[nOffset], &m_pbData[m_nOffset], m_nSize);
		m_nOffset = nOffset;
	}
#endif

	ASSERT_EX(*this, nSize + m_nOffset <= m_nCapacity);
	
	// Set new size
	m_nSize = nSize;
}


void CPacketData::Reserve(const arg_type nCapacity)
{
	// If this ASSERTions bombs, you passed a value that cannot be
	// held in a variable of type size_type
	ASSERT(nCapacity <= static_cast<size_type>(-1));
	
	// (Re-)allocate memory
	m_pbData = static_cast<unsigned char *>(realloc(m_pbData, 
		m_nCapacity = nCapacity));

	// If this ASSERTion bombs, we ran out of memory
	ASSERT_EX(*this, m_pbData);
	ASSERT_VALIDHEAPBLOCK(m_pbData, m_nCapacity);
}


void CPacketData::Append(const void *const pData, const arg_type cbData)
{
	// If this ASSERTions bombs, you passed a value that cannot be
	// held in a variable of type size_type
	ASSERT(cbData <= static_cast<size_type>(-1));	
	ASSERT(m_nSize + cbData <= static_cast<size_type>(-1));
	
#ifdef __CPACKET_ALLOW_GROW__
	// Allows packets to grow beyond the capacity reserved during 
	// construction
	if (m_nSize + cbData > m_nCapacity)
		Reserve(m_nSize + cbData);
#else
	// If you end up here, increase the inital capacity limit (header) for 
	// this packet, or allow packets to grow (not recommended for tiny 
	// devices)
	ASSERT_EX(*this, m_nSize + cbData <= m_nCapacity);
#endif
	
#ifdef __CPACKET_ALLOW_MOVE__
	// Move existing contents to make room for header
	if (cbData > m_nCapacity - m_nSize - m_nOffset)
	{
		const size_type nOffset = m_nCapacity - m_nSize - cbData;

		memmove(&m_pbData[nOffset], &m_pbData[m_nOffset], m_nSize);
		m_nOffset = nOffset;
	}
#endif

	// Fail, if not enough storage left
	ASSERT_EX(*this, m_nSize + m_nOffset + cbData <= m_nCapacity);
	
	// Append new data at the end
	memcpy(&m_pbData[m_nOffset + m_nSize], pData, cbData);
	m_nSize += cbData;
}


void CPacketData::Prepend(const void *const pData, const arg_type cbData)
{
	// If this ASSERTions bombs, you passed a value that cannot be
	// held in a variable of type size_type
	ASSERT(cbData <= static_cast<size_type>(-1));	
	ASSERT(m_nSize + cbData <= static_cast<size_type>(-1));

#ifdef __CPACKET_ALLOW_GROW__
	// Allows packets to grow beyond the capacity reserved during 
	// construction
	if (m_nSize + cbData > m_nCapacity)
		Reserve(m_nSize + cbData);
#else
	// If you end up here, increase the inital capacity limit (header) for 
	// this packet, or allow packets to grow (not recommended for tiny 
	// devices)
	ASSERT_EX(*this, m_nSize + cbData <= m_nCapacity);
#endif
	
#ifdef __CPACKET_ALLOW_MOVE__
	// Move existing contents to make room for header
	if (cbData > m_nOffset)
	{
		memmove(&m_pbData[cbData], &m_pbData[m_nOffset], m_nSize);
		m_nOffset = cbData;
	}
#endif

	// Fail, if not enough room available
	ASSERT_EX(*this, cbData <= m_nOffset);

	// Write header to front of packet
	memcpy(&m_pbData[m_nOffset -= cbData], pData, cbData);
	m_nSize += cbData;
}


void CPacketData::StripHeader(const arg_type cbHeader)
{
	// If this ASSERTions bombs, you passed a value that cannot be
	// held in a variable of type size_type
	ASSERT(cbHeader <= static_cast<size_type>(-1));	
	
	// Cannot remove more bytes than actually stored
	ASSERT_EX(*this, cbHeader <= m_nSize);
	
	// Increase offset to point to new data block, decrease size
	m_nOffset += cbHeader;
	m_nSize -= cbHeader;
}


void CPacketData::RemoveHeader(void *pHeader, const arg_type cbHeader)
{
	// If this ASSERTions bombs, you passed a value that cannot be
	// held in a variable of type size_type
	ASSERT(cbHeader <= static_cast<size_type>(-1));	
	
	// Cannot remove more bytes than actually stored
	ASSERT_EX(*this, cbHeader <= m_nSize);
	
	// Copy header...
	memcpy(pHeader, &m_pbData[m_nOffset], cbHeader);
	
	// Strip-off header
	StripHeader(cbHeader);
}


void CPacketData::StripTrailer(const arg_type cbTrailer)
{
	// If this ASSERTions bombs, you passed a value that cannot be
	// held in a variable of type size_type
	ASSERT(cbTrailer <= static_cast<size_type>(-1));
	
	// Cannot remove more bytes than actually stored
	ASSERT_EX(*this, cbTrailer <= m_nSize);
	
	// Shrink packet
	m_nSize -= cbTrailer;
}


void CPacketData::RemoveTrailer(void *pTrailer, const arg_type cbTrailer)
{
	// If this ASSERTions bombs, you passed a value that cannot be
	// held in a variable of type size_type
	ASSERT(cbTrailer <= static_cast<size_type>(-1));	
	
	// Cannot remove more bytes than actually stored
	ASSERT_EX(*this, cbTrailer <= m_nSize);
	
	// Copy trailer...
	memcpy(pTrailer, &m_pbData[m_nOffset + m_nSize - cbTrailer], 
		cbTrailer);
	
	// Strip-off trailer
	StripTrailer(cbTrailer);
}


/////////////////////////////////////////////////////////////////////////////
// CPacket

CPacket::CPacket(const CPacketData::arg_type nBodyAndTrailerCapacity, 
	const CPacketData::arg_type nHeaderCapacity) : m_pData(0)
{
	// This constructor creates a new, empty packet with the specified 
	// initial capacities for header, body and trailer
	Attach(new CPacketData(nBodyAndTrailerCapacity, nHeaderCapacity));
}


CPacket::CPacket(const void *const pData, const CPacketData::arg_type nSize,
	const CPacketData::arg_type nTrailerCapacity,
	const CPacketData::arg_type nHeaderCapacity) : m_pData(0)
{
	// Creates a new packet initialized with the specified body. Create a 
	// data object with enough space for body size plus trailer and header
	Attach(new CPacketData(nSize + nTrailerCapacity, nHeaderCapacity));

	// We have initial data. Use Append() to copy the data into the packet
	m_pData->Append(pData, nSize);
}


CPacket::CPacket(const CPacket &packet) : m_pData(packet.m_pData)
{
	// Tell data object, that it is referenced by another object
	if (m_pData)
		m_pData->AddReference();
}


CPacket::CPacket() : m_pData(0)
{
}


CPacket::~CPacket()
{
	// Release the CPacketData object. Do not use operator delete, instead 
	// call CPacketData::Release()
	if (m_pData)
	{
		ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
		m_pData->Release();
	}
}


CPacket CPacket::Attach(CPacketData *const pNewPacketData)
{
	CPacket packet = *this;

	// Hands-off, invalidate link to data object instance
	if (m_pData)
	{
		ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
		m_pData->Release();
	}

	// Increment reference counter on new CPacketData object, if any	
	if (pNewPacketData)
	{
		ASSERT_VALIDHEAPBLOCK(pNewPacketData, sizeof(CPacketData));
		pNewPacketData->AddReference();
	}

	m_pData = pNewPacketData;

	return packet;
}


void CPacket::SetSize(const CPacketData::arg_type nSize)
{
	ASSERT_EX(*this, m_pData);
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));

	// Prepare write access (copy-on-write)
	PrepareWriteAccess();

	m_pData->SetSize(nSize);
}


void CPacket::PrepareWriteAccess()
{
	// If this is not the only CPacket instance referencing the CPacketData
	// object, create a new copy of the data on the fly. This should be 
	// avoided whenever possible
	ASSERT_EX(*this, m_pData);	
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));

	if (m_pData->m_nReference > 1)
	{

#ifdef __CPACKET_ALLOW_COW__
		CPacketData *pData = m_pData;
		m_pData = new CPacketData(*pData);
		pData->Release();
#else
		// If you end up here, the copy-on-write feature has been disabled
		// and a write attempt on a shared packet has happened
		ASSERT_EX(*this, false);
#endif
		
	}
}


#ifdef _DEBUG

CPacketData::size_type CPacket::GetSize() const
{ 
	ASSERT_EX(*this, m_pData);
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
	return m_pData->GetSize(); 
}


CPacketData::size_type CPacket::GetCapacity() const
{ 
	ASSERT_EX(*this, m_pData);
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
	return m_pData->GetCapacity();
}


const void *CPacket::GetData() const
{ 
	ASSERT_EX(*this, m_pData); 
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
	return m_pData->GetData(); 
}


void *CPacket::GetData()
{ 
	ASSERT_EX(*this, m_pData); 
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
	PrepareWriteAccess(); 
	return m_pData->GetData(); 
}


const void *CPacket::GetDataAt(const CPacketData::arg_type nOffset) const
{ 
	// If this ASSERTions bombs, you passed a value that cannot be
	// held in a variable of type size_type
	ASSERT(nOffset <= static_cast<CPacketData::size_type>(-1));
	ASSERT_EX(*this, nOffset < GetCapacity());
	// ASSERT_EX(*this, nOffset < GetSize());
	
	return static_cast<const unsigned char *>(GetData()) + nOffset; 
}


void *CPacket::GetDataAt(const CPacketData::arg_type nOffset)
{
	// If this ASSERTions bombs, you passed a value that cannot be
	// held in a variable of type size_type
	ASSERT(nOffset <= static_cast<CPacketData::size_type>(-1));
	ASSERT_EX(*this, nOffset < GetCapacity());
	// ASSERT_EX(*this, nOffset < GetSize());
	
	return static_cast<unsigned char *>(GetData()) + nOffset; 
}

#endif


void CPacket::Append(const void *const pData, 
	const CPacketData::arg_type cbData)
{
	ASSERT_EX(*this, m_pData);
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
	
	// Prepare write access (copy-on-write)
	PrepareWriteAccess();

	m_pData->Append(pData, cbData);
}


void CPacket::Prepend(const void *const pData, 
	const CPacketData::arg_type cbData)
{
	ASSERT_EX(*this, m_pData);
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
	
	// Prepare write access (copy-on-write)
	PrepareWriteAccess();
	
	m_pData->Prepend(pData, cbData);
}


void CPacket::StripHeader(const CPacketData::arg_type cbHeader)
{
	ASSERT_EX(*this, m_pData);
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
	
	// Prepare write access (copy-on-write)
	PrepareWriteAccess();
	
	m_pData->StripHeader(cbHeader);
}


void CPacket::RemoveHeader(void *pHeader, 
	const CPacketData::arg_type cbHeader)
{
	ASSERT_EX(*this, m_pData);
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
		
	// Prepare write access (copy-on-write)
	PrepareWriteAccess();

	m_pData->RemoveHeader(pHeader, cbHeader);
}


void CPacket::StripTrailer(const CPacketData::arg_type cbTrailer)
{
	ASSERT_EX(*this, m_pData);
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
	
	// Prepare write access (copy-on-write)
	PrepareWriteAccess();

	m_pData->StripTrailer(cbTrailer);
}


void CPacket::RemoveTrailer(void *pTrailer, 
	const CPacketData::arg_type cbTrailer)
{
	ASSERT_EX(*this, m_pData);
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
	
	// Prepare write access (copy-on-write)
	PrepareWriteAccess();

	m_pData->RemoveTrailer(pTrailer, cbTrailer);
}


void CPacket::Clear(const CPacketData::arg_type nHeaderCapacity)
{
	ASSERT_EX(*this, m_pData);
	ASSERT_VALIDHEAPBLOCK(m_pData, sizeof(CPacketData));
	
	// Prepare write access (copy-on-write)
	PrepareWriteAccess();
	
	// Clear the entire packet
	m_pData->m_nOffset = nHeaderCapacity;
	m_pData->m_nSize = 0;	
}


CPacket CPacket::CreateCopy() const
{
	return IsEmpty() ? CPacket() : CPacket(GetData(), GetSize());
}


bool CPacket::operator==(const CPacket &packet) const
{
	// Equal if sizes match and data identical (or both packets empty)
	return (GetSizeEx() == packet.GetSizeEx()) && 
		!memcmp(GetData(), packet.GetData(), GetSizeEx());
}


bool CPacket::operator!=(const CPacket &packet) const
{
	return !(packet == *this);
}


CPacket &CPacket::operator=(const CPacket &packet)
{
	Attach(packet.m_pData);
	
	return *this;
}
