// Packet.h : header file
//
// Copyright(C) 2008-2010 ubisys technologies GmbH, Duesseldorf, Germany.
// All rights reserved.
//
// www.ubisys.de
// support@ubisys.de
//
// Part of the ubisys(R) Compact15.4(TM) class library
//
// A powerful packet container class that provides easy packet assembly and 
// disassembly in an efficient manner. This implementation allows the stack
// implementor to reserve space before and after the body. This allows adding
// headers and trailers to the actual payload as the packet traverses the 
// stack. If enough space is reserved, no memory moving or copying is 
// required. A copy-on-write feature further improves the overall design.
// The implementation is scalable. The stack designer may enable/disable 
// features individually depending on the runtime environment (e.g. RFD vs. 
// FFD device, tiny active RFID tag vs. PC, etc.).
//
// __CPACKET_ALLOW_GROW__ enables automatic packet growth beyond initial 
// capacity limits (within capacity limits, automatic packet growth is always
// possible). __CPACKET_ALLOW_MOVE__ must be defined, too.
//
// __CPACKET_ALLOW_MOVE__ allows moving packet contents within the overall 
// reserved space. This allows to shift between header and trailer capacity
// as required at the expence of a call to memmove().
//
// #define __CPACKET_ALLOW_COW__, to allow copy-on-write on packet level. 
// Does not require __CPACKET_ALLOW_GROW__
//
// __CPACKET_SIZETYPE__ defines the data type used to store sizes, offsets,
// and capacity limits (default is byte yielding a maximum capacity of 255 
// bytes). For 64KB capacity use unsigned short, for 4GB use unsigned long

/////////////////////////////////////////////////////////////////////////////
// Default CPacketData::size_type, if none is defined

#ifndef __CPACKET_SIZETYPE__
#define __CPACKET_SIZETYPE__ unsigned char
#endif


/////////////////////////////////////////////////////////////////////////////
// CPacketData

class CPacketData
{
	friend class CPacket;

	// typedefs & enums
	public:
		// Type of size, offset, and capacity members
		typedef __CPACKET_SIZETYPE__ size_type;
		
		// Type of size, offset, and capacity arguments
		typedef size_t arg_type;

	// Construction
	public:
		explicit CPacketData(const arg_type nBodyAndTrailerCapacity, 
			const arg_type nHeaderCapacity = 0);

		CPacketData(const CPacketData &data);

	// Operations
	public:
		// Increments the reference count by one and returns its current 
		// value
		size_type AddReference();

		// Decrements the reference count by one. If it reaches zero, the 
		// object is destroyed. As soon as Release() has returned zero, the
		// instance has been deleted and must not be accessed any more
		size_type Release();

		// Appends data at the end of the packet
		void Append(const void *const pData, const arg_type cbData);

		// Inserts data in front of the packet
		void Prepend(const void *const pData, const arg_type cbData);

		// Strips off a header, i.e. removes it from the packet
		void StripHeader(const arg_type cbHeader);

		// Extracts a header and removes it from the packet
		void RemoveHeader(void *pHeader, const arg_type cbHeader);

		// Strips of a trailer, i.e. removes it from the packet
		void StripTrailer(const arg_type cbTrailer);

		// Extracts a trailer and removes it from the packet
		void RemoveTrailer(void *pTrailer, const arg_type cbTrailer);

		// Returns the length of the packet
		size_type GetSize() const;
		
		// Returns the number of bytes that can be maximally stored in this
		// packet
		size_type GetCapacity() const;

		// Sets the size of the packet
		void SetSize(const arg_type nSize);

		// Returns a pointer to the start of packet data
		void *GetData() const;

	// Implementation
	public:
		~CPacketData();
		
#ifdef _DEBUG
		// Dumps the packet data
		void Dump(const arg_type nStart = 0, const bool bAsHex = true, 
			const bool bAsText = false) const;
#endif

	protected:
		// Packet data of variable length
		unsigned char *m_pbData;
		
		// Number of bytes currently stored in the body section
		size_type m_nSize;
		
		// Number of bytes that can be maximally stored in this packet
		size_type m_nCapacity;

		// Indicates the beginning of the actual data payload. Allows reserving
		// space in front of the payload for headers added as the packet 
		// traverses the protocol stack
		size_type m_nOffset;

		// Reference counter. Each CPacket instance referencing this object 
		// increments the reference count by one
		size_type m_nReference;

		// Allocates memory large enough to store at least nCapacity elements
		void Reserve(const arg_type nCapacity);
};

inline void *CPacketData::GetData() const
	{ return m_pbData + m_nOffset; }

inline CPacketData::size_type CPacketData::GetSize() const
	{ return m_nSize; }

#ifdef __CPACKET_ALLOW_MOVE__
inline CPacketData::size_type CPacketData::GetCapacity() const
	{ return m_nCapacity; }
#else
inline CPacketData::size_type CPacketData::GetCapacity() const
	{ return m_nCapacity - m_nOffset; }
#endif


/////////////////////////////////////////////////////////////////////////////
// CPacket

class CPacket
{
	// Construction
	public:
		// Creates a new, empty packet with the specified initial capacities
		explicit CPacket(const CPacketData::arg_type nBodyAndTrailerCapacity, 
			const CPacketData::arg_type nHeaderCapacity = 0);

		// Creates a new packet initialized with the specified body. 
		// Additional space is reserved for header and trailer
		explicit CPacket(const void *const pData, 
			const CPacketData::arg_type nSize,
			const CPacketData::arg_type nTrailerCapacity = 0,
			const CPacketData::arg_type nHeaderCapacity = 0);

		// Copy constructor
		CPacket(const CPacket &packet);
		
		// Creates an empty packet object (without associated data)
		CPacket();
		
	// Attributes
	public:
		// Returns true, if the packet is empty, i.e. no packet data has been
		// attached to this object
		bool IsEmpty() const;		
		
		// Returns the length of the packet. Call this version if you know that
		// packet data object has been attachted to this object, otherwise call
		// GetSizeEx(), which additionally checks against an empty packet
		CPacketData::size_type GetSize() const;
		
		// Returns the length of the packet. Call this version if you do not 
		// know whether a packet data object has been attachted to this object, 
		// otherwise call GetSize() for improved performance
		CPacketData::size_type GetSizeEx() const;
				
		// Returns the number of bytes that can be maximally stored
		CPacketData::size_type GetCapacity() const;		

		// Sets the size of the packet
		void SetSize(const CPacketData::arg_type nSize);

		// Returns a read-only pointer to packet data.
		const void *GetData() const;
		
		// Returns a read-only pointer to a location within packet data
		const void *GetDataAt(const CPacketData::arg_type nOffset) const;

		// Returns a writable pointer to packet data. Use the read-only 
		// version if you do not really need to write to the packet
		void *GetData();

		// Returns a writable pointer to a location within packet data. Use the 
		// read-only version if you do not really need to write to the packet
		void *GetDataAt(const CPacketData::arg_type nOffset);

	// Operations
	public:
		// Appends data at the end of the packet
		void Append(const void *const pData, 
			const CPacketData::arg_type cbData);
		
		// Appends data stored in another packet at the end of this packet
		void Append(const CPacket &packet);

		// Inserts data in front of the packet
		void Prepend(const void *const pData, 
			const CPacketData::arg_type cbData);

		// Inserts data stored in another packet in front of this packet
		void Prepend(const CPacket &packet);
		
		// Strips of a header, i.e. removes it from the packet
		void StripHeader(const CPacketData::arg_type cbHeader);

		// Extracts a header and removes it from the packet
		void RemoveHeader(void *pHeader, 
			const CPacketData::arg_type cbHeader);

		// Strips of a trailer, i.e. removes it from the packet
		void StripTrailer(const CPacketData::arg_type cbTrailer);

		// Extracts a trailer and removes it from the packet
		void RemoveTrailer(void *pTrailer, 
			const CPacketData::arg_type cbTrailer);

		// Clears the packet contents, such that the packet provides its 
		// specified capacity. Header capacity may be specified if non-zero
		void Clear(const CPacketData::arg_type nHeaderCapacity = 0);
		
		// True, if packets are equal
		bool operator==(const CPacket &packet) const;

		// True, if packets are unequal
		bool operator!=(const CPacket &packet) const;

		// Assignment operator		
		CPacket &operator=(const CPacket &packet);

		// Detaches the underlying packet object container and returns a packet 
		// object that can be passed to higher/lower layers. Then attaches 
		// another container object (or a null object). By using Attach() the 
		// caller declares that no further access is intended to the original
		// packet object. If a null pointer is passed, the CPacketData object 
		// will be invalidated, i.e. access to packet data is prohibited
		CPacket Attach(CPacketData *const pNewPacketData = 0);

		// Calls Attach() with a null object
		CPacket Detach();
		
		// Explicitely creates a copy (also works, if copy-on-write is not
		// enabled). Should only be used when a true copy is required
		CPacket CreateCopy() const;
		
#ifdef _DEBUG
		void Dump(const TCHAR *const pszMessage = 0, 
			const CPacketData::arg_type nStart = 0,
			const bool bAsHex = true, const bool bAsText = false) const;
#endif

	// Implementation
	public:		
		virtual ~CPacket();

	protected:
		// The CPacketData object that holds the actual data
		CPacketData *m_pData;

		// Called internally whenever write access to packet data is 
		// required. Performs copy-on-write if neccessary
		void PrepareWriteAccess();
};

#ifndef _DEBUG

inline CPacketData::size_type CPacket::GetSize() const
	{ return m_pData->GetSize(); }

inline CPacketData::size_type CPacket::GetCapacity() const
	{ return m_pData->GetCapacity(); }

inline const void *CPacket::GetData() const
	{ return m_pData->GetData(); }

inline void *CPacket::GetData()
	{ PrepareWriteAccess(); return m_pData->GetData(); }

inline const void *CPacket::GetDataAt(const CPacketData::arg_type nOffset) const
	{ return static_cast<const unsigned char *>(GetData()) + nOffset; }

inline void *CPacket::GetDataAt(const CPacketData::arg_type nOffset)
	{ return static_cast<unsigned char *>(GetData()) + nOffset; }

#endif

inline bool CPacket::IsEmpty() const
	{ return m_pData ? false : true; }

inline CPacketData::size_type CPacket::GetSizeEx() const
	{ return IsEmpty() ? 0 : GetSize(); }

inline void CPacket::Append(const CPacket &packet) 
	{ Append(packet.GetData(), packet.GetSize()); }
				
inline void CPacket::Prepend(const CPacket &packet) 
	{ Prepend(packet.GetData(), packet.GetSize()); }

inline CPacket CPacket::Detach()
	{ return Attach(); }
