// SerialCommandInterface.cpp : implementation file
//
// Copyright(C) 2009 ubisys technologies GmbH, Duesseldorf, Germany.
// All rights reserved.
//
// www.ubisys.de
// support@ubisys.de
//
// THIS SOURCE CODE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//
// IN NO EVENT SHALL UBISYS TECHNOLOGIES BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOURCE CODE OR THE USE OR
// OTHER DEALINGS IN THE SOURCE CODE.
//
// Part of the ubisys(R) 13.56 MHz RFID USB stick demo application
//
// Provides simple high-level functions to control a ubisys RFID USB module
// via its AT like command set

#include <pthread.h>
#include <fcntl.h>
#include <sys/time.h>
#include <termios.h>
#include <cstdlib>
#include <csignal>
#include <cstring>
#include <cstdarg>
#include <cerrno>
#include <vector>
#include <string>
#include <list>
#include <sstream>
#include "CompactFramework.h"
#include "Packet.h"
#include "CommunicationsQueue.h"
#include "CommQueue.h"
#include "SerialCommandInterface.h"

using namespace std;

/////////////////////////////////////////////////////////////////////////////
// CSerialApplicationInterface

CSerialCommandInterface::CSerialCommandInterface(CCommQueue &queue) :
m_queue(queue), m_bAbort(false)
{
	// Setup queue
	m_queue.Initialize();
	m_queue.EnableTransmission();
	m_queue.EnableReception();
}


void CSerialCommandInterface::AbortReception()
{
	m_bAbort = true;
	pthread_cond_broadcast(&m_queue.m_conditionReceive);
}


bool CSerialCommandInterface::SendRequest(CPacket &request)
{

#ifdef _DEBUG
	request.Dump("request");
#endif

	// Transmit the request frame
	m_queue.Send(request);

	return true;
}


bool CSerialCommandInterface::ReceiveResponse(CPacket &response, 
	const unsigned int nTimeout, bool *const pbAborted)
{
	ASSERT(response.IsEmpty());
	
	m_bAbort = false;
	
	while (!m_bAbort && !response.GetSizeEx())
	{
		struct timespec timeout;

		if (nTimeout)
		{
			struct timeval time;
			
			gettimeofday(&time, 0);
			timeout.tv_sec = time.tv_sec;
			timeout.tv_nsec = time.tv_usec * 1000u + nTimeout * 1000000u;
			
			if (timeout.tv_nsec > 1000000000ll)
			{
				timeout.tv_sec += timeout.tv_nsec / 1000000000ll;
				timeout.tv_nsec %= 1000000000ll;
			}
		}
			
		pthread_mutex_lock(&m_queue.m_csr);
				
		const int nResult = nTimeout ? 
			pthread_cond_timedwait(&m_queue.m_conditionReceive, 
				&m_queue.m_csr, &timeout) :
			pthread_cond_wait(&m_queue.m_conditionReceive, &m_queue.m_csr);

		if (nResult || m_bAbort)
		{
			TRACE2("SCI: no response (%d, %saborted).\r\n", nResult,
				m_bAbort ? "" : "not ");
				
			if (pbAborted)
				*pbAborted = m_bAbort;
				
			m_bAbort = false;
			
			return false;
		}
			
		// Upon successful wait the mutex is locked, so unlock first...
		pthread_mutex_unlock(&m_queue.m_csr);
			
		// Return true
		m_queue.Receive(response);
				
#ifdef _DEBUG
		response.Dump("response");
#endif
	
		for (CPacketData::size_type i = 0; i < response.GetSizeEx(); i++)
			TRACE1("%c", *static_cast<char *>(response.GetDataAt(i)));
	}

	if (pbAborted)
		*pbAborted = m_bAbort;

	m_bAbort = false;

	return true;
}


bool CSerialCommandInterface::SplitResponse(const CPacket &response, 
	std::vector<std::string> &lines)
{
	if (!response.GetSizeEx())
		return false;

	ASSERT(lines.empty());

	std::string strLine;
	
	for (unsigned int i = 0; i < response.GetSize(); i++)
	{
		const char ch = *static_cast<const char *>(response.GetDataAt(i));

		if (ch == '\n')
		{
			lines.push_back(strLine);
			strLine.clear();
		}
		else
			strLine += ch;
	}

	if (!strLine.empty())
		lines.push_back(strLine);

	return !lines.empty();
}


string CSerialCommandInterface::TrimLeft(const string &str,
	const char *const pszDelimiters)
{
	string::size_type startpos = str.find_first_not_of(pszDelimiters);
	
	return (startpos == string::npos) ? str : str.substr(startpos);
}		


string CSerialCommandInterface::TrimRight(const string &str,
	const char *const pszDelimiters)
{
	string::size_type endpos = str.find_last_not_of(pszDelimiters);
	
	return (endpos == string::npos) ? str : str.substr(0, endpos + 1);
}		


string CSerialCommandInterface::Trim(const string &str,
	const char *const pszDelimiters)
{
	return TrimLeft(TrimRight(str, pszDelimiters), pszDelimiters);
}


string CSerialCommandInterface::Tokenize(const string &str,
	const char *const pszDelimiters, string::size_type &start)
{
	// Skip over leading delimiters
	start = str.find_first_not_of(pszDelimiters, start);
	
	if (start == string::npos)
		return "";
		
	string::size_type stop = str.find_first_of(pszDelimiters, start);
	
	const string result = (stop == string::npos) ? 
		str.substr(start) : 
		str.substr(start, stop - start);
	
	start = stop + 1;
	
	return result;	
}		


string CSerialCommandInterface::SpanExcluding(const string &str,
	const char *const pszDelimiters)
{
	// Skip over leading delimiters
	string::size_type stop = str.find_first_of(pszDelimiters);
	
	return (stop == string::npos) ? str : str.substr(0, stop);
}		


bool CSerialCommandInterface::VerifyResponse(const std::vector<std::string> &response)
{
	if (response.empty())
		return false;

	return response[response.size() - 1] == "OK\r";
}


bool CSerialCommandInterface::ExchangeCommand(CPacket &request, 
	CPacket &response, const unsigned int nTimeout)
{
	// Transmit command and receive response
	return SendRequest(request) && ReceiveResponse(response, nTimeout);
}


bool CSerialCommandInterface::ExecuteEx(CPacket &request,
	std::vector<std::string> *pResponse, const unsigned int nTimeout, 
	CPacket *pRawResponse)
{
	CPacket response;
	std::vector<std::string> lines;

	// Transmit command and receive response
	if (!ExchangeCommand(request, response, nTimeout) ||
		!SplitResponse(response, lines) || !VerifyResponse(lines))
		return false;

	if (pResponse)
		pResponse->assign(lines.begin(), lines.end());

	if (pRawResponse)
		*pRawResponse = response;

	return true;
}


bool CSerialCommandInterface::Execute(const char *const pszRequest, 
	std::vector<std::string> *pResponse, const unsigned int nTimeout, 
	CPacket *pRawResponse)
{
	CPacket request(pszRequest, strlen(pszRequest), 1), response;
	static const char chCarriageReturn = '\r';

	ASSERT(pszRequest[strlen(pszRequest) - 1] != chCarriageReturn);
	request.Append(&chCarriageReturn, sizeof(char));

	return ExecuteEx(request, pResponse, nTimeout, pRawResponse);
}


bool CSerialCommandInterface::EnableEcho(const bool bEnable)
{
	ostringstream s;
	
	s << "ATE" << (bEnable ? '1' : '0');

	// Send request frame, wait for response, and verify it
	return Execute(s.str().c_str(), 0, 100);
}


bool CSerialCommandInterface::GetVersion(std::string &strVersion, 
	std::string &strSerialNumber)
{
	std::vector<std::string> response;
	
	// Send identify request command, and wait for response
	if (!Execute("ATI", &response) || response.size() != 3)
		return false;
		
	strVersion = TrimRight(response[0], "\r");
	strSerialNumber = TrimRight(response[1], "\r");

	return true;
}


bool CSerialCommandInterface::PrepareFlashUpdate()
{
	// Send identify request command, and wait for response
	Execute("AT+flash-update!");

	return true;
}


bool CSerialCommandInterface::EnableField(const bool bEnable)
{
	ostringstream s;
	
	s << "AT+RF=" << (bEnable ? '1' : '0');

	// Send request frame, wait for response, and verify it
	return Execute(s.str().c_str());
}


bool CSerialCommandInterface::QueryField(bool &bEnabled)
{
	std::vector<std::string> response;

	// Send query RF command, and wait for response
	if (!Execute("AT+RF?", &response) || response.size() != 2)
		return false;

	if ((response[0].length() != 6) || 
		(response[0].substr(0, 4) != "+RF:") ||
		((response[0][4] != '0') && (response[0][4] != '1')))
		return false;
	
	bEnabled = response[0][4] == '1';

	return true;
}


bool CSerialCommandInterface::EnableAutoScan(const bool bEnable)
{
	ostringstream s;

	s << "AT+SCAN" << (bEnable ? '1' : '0');

	// Send request frame, wait for response, and verify it
	return Execute(s.str().c_str());
}


bool CSerialCommandInterface::AnticipateScanResults(const RECEIVECALLBACK pfnReceive, 
	void *const pArgument, const unsigned int nTimeout)
{
	bool bAborted = false;
	
	while (!bAborted)
	{
		CPacket response;
		
		if (ReceiveResponse(response, nTimeout, &bAborted))
		{
			// Accept board response... 
			std::vector<std::string> lines;

			// Invoke the caller-supplied notification
			if (!SplitResponse(response, lines))
				return false;

			if (!pfnReceive(*this, lines, pArgument))
				return false;
		}
	}
	
	return true;
}


bool CSerialCommandInterface::RetrieveUniqueID(const std::string &strUniqueID,
	unsigned long long &nUniqueID)
{
	char *pszEnd;

	nUniqueID = strtoull(strUniqueID.c_str(), &pszEnd, 16);

	// Convert from ASCII hex to unsigned 64-bit
	if (pszEnd != strUniqueID.c_str() + strUniqueID.length())
		return false;

	// For ISO15693, all tag IDs must start with E0...
	if (nUniqueID >> 56 != 0xe0)
		return false;

	return true;
}


bool CSerialCommandInterface::Inventory(std::list<unsigned long long> &uids)
{
	std::vector<std::string> response;

	// Send inventory command, and wait for response
	if (!Execute("AT+I", &response))
		return false;

	// Response: +UID=E0040100005887FC (possibly multiple lines)
	uids.clear();

	for (std::vector<std::string>::size_type i = 0; i < response.size() - 1; i++)
	{
		unsigned long long nUniqueID;

		if ((response[i].length() < 31) || 
			(response[i].substr(0, 5) != "+UID="))
			return false;

		if (!RetrieveUniqueID(response[i].substr(5, 16), nUniqueID))
			return false;

		uids.push_back(nUniqueID);
	}

	return true;
}


bool CSerialCommandInterface::GetSystemInformation(unsigned long long &nUID,
	unsigned char &bFlags, unsigned char &bDSFID, unsigned char &bAFI, 
	unsigned char &bBlockCount, unsigned char &bBlockSize,
	unsigned char &bICReference)
{
	std::vector<std::string> response;

	// Send "Get System Information" command, and wait for response
	if (!Execute("AT+S", &response))
		return false;

	// Response: +UID=E0040100005887FC,DSFID=00,AFI=01,BC=28,BS=4,IC=01,
	// where only UID is mandatory, other fields are optional

	string::size_type nPosition = 0;
	string strToken = Tokenize(response[0], ",\r", nPosition);

	if ((strToken.length() != 21) || (strToken.substr(0, 5) != "+UID="))
		return false;

	if (!RetrieveUniqueID(strToken.substr(5, 16), nUID))
		return false;

	bFlags = 0;

	// Process optional fields
	while (!(strToken = Tokenize(response[0], ",\r", nPosition)).empty())
	{
		string::size_type nToken = 0;
		string strKey = Tokenize(strToken, "=", nToken);
		string strValue = Tokenize(strToken, "=", nToken);
		char *pszEnd;
		
		if (strKey.empty() || strValue.empty())
			return false;
		
		if (strKey == "DSFID")
		{
			bDSFID = static_cast<unsigned char>(strtoul(strValue.c_str(), &pszEnd, 16));

			// Convert from ASCII hex to unsigned 64-bit
			if (pszEnd == strValue.c_str() + strValue.length())
				bFlags |= infoDSFID;
		}
		else if (strKey == "AFI")
		{
			bAFI = static_cast<unsigned char>(strtoul(strValue.c_str(), &pszEnd, 16));

			// Convert from ASCII hex to unsigned 64-bit
			if (pszEnd == strValue.c_str() + strValue.length())
				bFlags |= infoAFI;
		}
		else if (strKey == "BC")
		{
			bBlockCount = static_cast<unsigned char>(strtoul(strValue.c_str(), &pszEnd, 10));

			// Convert from ASCII decimal to unsigned 64-bit
			if (pszEnd == strValue.c_str() + strValue.length())
				bFlags |= infoVICCMemorySize;
		}
		else if (strKey == "BS")
		{
			bBlockSize = static_cast<unsigned char>(strtoul(strValue.c_str(), &pszEnd, 10));

			// Convert from ASCII hex to unsigned 64-bit
			if (pszEnd == strValue.c_str() + strValue.length())
				bFlags |= infoVICCMemorySize;
		}
		else if (strKey == "IC")
		{
			bICReference = static_cast<unsigned char>(strtoul(strValue.c_str(), &pszEnd, 16));

			// Convert from ASCII hex to unsigned 64-bit
			if (pszEnd == strValue.c_str() + strValue.length())
				bFlags |= infoICReference;
		}
		else
			return false;
	}

	return true;
}


bool CSerialCommandInterface::ReadSingleBlock(const unsigned char nBlock,
	CPacket &data)
{
	std::vector<std::string> response;
	ostringstream s;
	CPacket raw;

	s << "AT+R" << static_cast<unsigned int>(nBlock);

	// Send "Read Single Block" command, and wait for response
	if (!Execute(s.str().c_str(), &response, 500, &raw))
		return false;

	// Response format example:
	// +DATA 4:ABCD
	// OK

	// Array of response lines is guaranteed not to be empty at this point. 
	// So it is safe to compare...
	if ((response[0].length() < 8) || (response[0].substr(0, 6) != "+DATA "))
		return false;

	string strSize = SpanExcluding(response[0].substr(6), ":");
	char *pszEnd;
		
	// Convert from ASCII decimal to int
	size_t cbData = strtoul(strSize.c_str(), &pszEnd, 10);

	if (pszEnd != strSize.c_str() + strSize.length())
		return false;

	// ISO15693: A single block can have no more than 256 bits = 32 bytes
	if (cbData > 32u)
		return false;

	ASSERT(response[0][6 + strSize.length()] == ':');

	const size_t nDataStart = 7 + strSize.length();

	// Data length match? (account for carriage-return)
	if (raw.GetSize() < static_cast<CPacketData::size_type>(nDataStart + cbData))
		return false;

	// Retrieve data (make a copy of the raw data portion)
	data = CPacket(raw.GetDataAt(nDataStart), cbData);

	return true;
}


bool CSerialCommandInterface::WriteSingleBlock(const unsigned char nBlock,
	const CPacket &data, const bool bOptionFlag)
{
	ostringstream s;

	// Check the data length. Notice that this is only a coarse check. We do 
	// not check if the data length matches the transponder's memory block 
	// size here. It is left to the caller to ensure that transponder block 
	// size and specified data length match exactly (either through a call
	// to GetSystemInformation() or a priori knowledge of the correct size)
	if (data.IsEmpty() || (data.GetSize() > 32))
		return false;

	s << "AT+" << (bOptionFlag ? 'W' : 'w') << static_cast<unsigned int>(nBlock) << ",";
	s << static_cast<unsigned int>(data.GetSize()) << '\r';

	// Create data packet, leaving room for memory block contents
	CPacket command(s.str().c_str(), s.str().length(), data.GetSize());

	// Append data to write
	command.Append(data);

	// ... followed by data to write, and wait for response
	return ExecuteEx(command);
}
