// main.cpp : implementation file
//
// Copyright(C) 2009 ubisys technologies GmbH, Duesseldorf, Germany.
// All rights reserved.
//
// www.ubisys.de
// support@ubisys.de
//
// THIS SOURCE CODE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
// EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF 
// MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
//
// IN NO EVENT SHALL UBISYS TECHNOLOGIES BE LIABLE FOR ANY CLAIM, DAMAGES OR
// OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, 
// ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOURCE CODE OR THE USE OR
// OTHER DEALINGS IN THE SOURCE CODE.
//
// ubisys(R) 13.56 MHz RFID USB stick demo application for Linux

#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/errno.h>
#include <pthread.h>
#include <termios.h>
#include <cstdio>
#include <csignal>
#include <cstring>
#include <string>
#include <list>
#include <set>
#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <stdexcept>
#include "CompactFramework.h"
#include "Packet.h"
#include "CommunicationsQueue.h"
#include "CommQueue.h"
#include "SerialCommandInterface.h"

using namespace std;

/////////////////////////////////////////////////////////////////////////////
// CAutoScanThreadCookie

class CAutoScanThreadCookie
{
	// Construction
	public:
		CAutoScanThreadCookie(CSerialCommandInterface &sci);
		
	// Attributes
	public:
		// SCI instance
		CSerialCommandInterface &m_sci;
		
		// Set of UIDs currently in the field
		set<unsigned long long> m_uids;
};

CAutoScanThreadCookie::CAutoScanThreadCookie(CSerialCommandInterface &sci) :
m_sci(sci)
{
}


/////////////////////////////////////////////////////////////////////////////
// OnIndicateScanResult()
//
// This call-back function is invoked by AnticipateScanResults() whenever the
// devices reports presence/loss of a tag. This function executes from auto-
// scan thread context

static bool OnIndicateScanResult(CSerialCommandInterface &sci,
	std::vector<std::string> &lines, void *pArgument)
{
	set<unsigned long long> &uids = 
		*static_cast<set<unsigned long long> *>(pArgument); 

	// Expect one or more strings like SCAN:+UID=E0040100005887FC, where
	// '+' indicates presence of a tag, '-' indicates loss of a tag
	for (std::vector<std::string>::size_type i = 0; i < lines.size(); i++)
	{
		if ((lines[i].length() < 26) || (lines[i].substr(0, 5) != "SCAN:") ||
			((lines[i][5] != '+') && (lines[i][5] != '-')))
			return false;

		unsigned long long nUniqueID;
		
		if (!CSerialCommandInterface::RetrieveUniqueID(lines[i].substr(10, 16), nUniqueID))
			return false;

		if (lines[i][5] == '+')
		{
			uids.insert(nUniqueID);
			cout << "Found tag " << hex << nUniqueID << endl;
		}
		else
		{
			uids.erase(nUniqueID);
			cout << "Lost tag " << hex << nUniqueID << endl;
		}
	}

	return true;
}


/////////////////////////////////////////////////////////////////////////////
// AutoScanThread()
//
// Thread procedure that runs the auto-scan in parallel to the normal program
// flow. Allows the main programm to stay responsive to user interaction, 
// while this thread watches out for transponders entering/leaving the field 

static void *AutoScanThread(void *pArgument)
{
	CAutoScanThreadCookie &cookie = 
		*static_cast<CAutoScanThreadCookie *>(pArgument);

	cout << "Entering auto-scan mode (separate thread)... ";

	// Enable auto-scan mode...	
	if (cookie.m_sci.EnableAutoScan())
		cout << " OK." << endl;
	else		
		cout << " failed." << endl;
		
	// ... and wait for incoming "SCAN:" indications
	cookie.m_sci.AnticipateScanResults(&OnIndicateScanResult, &cookie.m_uids);
	
	cout << "Terminating auto-scan thread." << endl;
	
	return 0;		
}


/////////////////////////////////////////////////////////////////////////////
// ShowVersion()

static void ShowVersion(CSerialCommandInterface &sci)
{
	string strVersion, strSerialNumber;
	
	if (sci.GetVersion(strVersion, strSerialNumber))
	{
		cout << "Successfully connected to:" << endl;
		cout << strVersion.c_str() << ", " << strSerialNumber << endl << endl;
	}
	else
		throw runtime_error("GetVersion() failed.");
}


/////////////////////////////////////////////////////////////////////////////
// DemonstrateInventory()

static void DemonstrateInventory(CSerialCommandInterface &sci)
{
	list<unsigned long long> uids;
		
	if (sci.Inventory(uids))
	{
		cout << "Inventory: found " << uids.size() << " tags:" << endl;
		
		if (!uids.empty())
			cout << hex << uids.front() << dec << endl;
	}
	else
		cout << "Inventory() failed." << endl;			
} 


/////////////////////////////////////////////////////////////////////////////
// DemonstrateSystemInformation()

static void DemonstrateSystemInformation(CSerialCommandInterface &sci)
{ 
	unsigned long long nUID;
	unsigned char bFlags, bDSFID, bAFI, bBlockCount, bBlockSize, bICReference;
		
	if (sci.GetSystemInformation(nUID, bFlags, bDSFID, bAFI, bBlockCount, bBlockSize, bICReference))
	{
		cout << "UID = " << hex << nUID << dec << endl;
		
		if (bFlags & CSerialCommandInterface::infoDSFID)
			cout << "DSFID = " << static_cast<unsigned int>(bDSFID) << endl;
			
		if (bFlags & CSerialCommandInterface::infoAFI)
			cout << "AFI = " << static_cast<unsigned int>(bAFI) << endl;
		
		if (bFlags & CSerialCommandInterface::infoVICCMemorySize)
		{
			const unsigned int nSize = static_cast<unsigned int>(bBlockCount) * 
				static_cast<unsigned int>(bBlockSize);
				
			cout << "On-chip memory: " << nSize << " bytes (" 
				<< static_cast<unsigned int>(bBlockCount) << " blocks, " 
				<< static_cast<unsigned int>(bBlockSize) << " bytes each)" 
				<< endl;
		}
		
		if (bFlags & CSerialCommandInterface::infoICReference) 
			cout << "IC reference = " << static_cast<unsigned int>(bICReference) << endl;
	}			
	else
		cout << "GetSystemInformation() failed." << endl;
}


/////////////////////////////////////////////////////////////////////////////
// DemonstrateMemoryAccess()

static void DemonstrateMemoryAccess(CSerialCommandInterface &sci, 
	const bool bWriteOptionFlag)
{
	CPacket data;
	
	cout << "Reading block 0..."; 
				
	if (sci.ReadSingleBlock(0, data))
	{

#ifdef _DEBUG
		data.Dump("block 0 (read)");
#endif

		cout << "OK:";
		
		// Print Hex-dump of memory block
		for (CPacketData::size_type i = 0; i < data.GetSize(); i++)
		{
			cout << ' ' << hex << setw(2) << setfill('0') 
				<< static_cast<unsigned int>(*static_cast<unsigned char *>(data.GetDataAt(i)));
		}

		cout << '.' << endl;

		cout << "Writing block 0 (inverted data)...";
		
		// Invert contents of memory block
		for (CPacketData::size_type i = 0; i < data.GetSize(); i++)
		{
			*static_cast<unsigned char *>(data.GetDataAt(i)) = 
				~*static_cast<unsigned char *>(data.GetDataAt(i));
		}
		 
		if (sci.WriteSingleBlock(0, data, bWriteOptionFlag))
		{
			cout << " OK." << endl;
				
#ifdef _DEBUG		
			data.Dump("block 0 (write)");
#endif
		}
		else
			cout << " failed." << endl;
	}
	else
		cout << " failed." << endl;
}


/////////////////////////////////////////////////////////////////////////////
// main()

int main(int argc, char *argv[])
{
	cout << "ubisys(R) 13.56 MHz RFID USB CDC/ACM Demo Application" << endl;
	cout << "CopyrightR) 2009 ubisys(R) technologies GmbH" << endl;
	cout << "All rights reserved." << endl << endl;
	
	try
	{
		CCommQueue queue;

		if (argc < 2)
		{
			cerr << "usage: rfidcdc <device>" << endl;
			throw runtime_error("invalid command line");
		}
		
		// Try to open the device (cdc_acm device node)		
		if (!queue.Open(argv[1]))
			throw runtime_error("Failed to open device");

		// Setup the device			
		if (!queue.Setup())
			throw runtime_error("Failed to setup device");

		// Associate a serial command interface protocol processor with the
		// communications queue (provides high-level AT interface support)			
		CSerialCommandInterface sci(queue);

		// Initially, disable echo and leave auto-scan mode (just in case it 
		// was enabled when this program started), then clear receive buffer
		sci.EnableEcho(false);
		sci.EnableAutoScan(false);
		queue.PurgeReceptionBuffer();

		// Display device type and firmware version
		ShowVersion(sci);
		
		// Perform single-shot inventory scan
		DemonstrateInventory(sci);
		
		// Demonstrate "Get System Information" request
		DemonstrateSystemInformation(sci);
	
		// Demonstrate read and write to transponder memory
		DemonstrateMemoryAccess(sci, (argc > 2) && !strcmp(argv[2], "W"));
		
		// BLOCK: Enter auto-scan mode under control of main thread...
		{
			CAutoScanThreadCookie cookie(sci);
			pthread_t hAutoScanThread;
			
			if (pthread_create(&hAutoScanThread, 0, &AutoScanThread, &cookie))
				cout << "Failed to create thread." << endl;
				
			// Notice: We entered auto-scan mode with above instruction. At
			// this point, we will be notified about tags entering or leaving
			// the RF field via OnIndicateScanResult(). 
				
			// Here could be any processing, like waiting for TCP/IP network 
			// connections, processing user input, etc.
			cout << "Press Enter to terminate..." << endl;
			cin.get();
			
			// To leave auto-scan mode (for example because the program is 
			// about to terminate), call AbortReception()...			
			sci.AbortReception();
			
			// Wait until the auto-scan thread has terminated
			cout << "Waiting for auto-scan thread to terminate..." << flush;
			pthread_join(hAutoScanThread, 0);				
			 
			cout << "Following transponders where in the field: " << endl;
			
			for (set<unsigned long long>::const_iterator i = cookie.m_uids.begin(); 
				 i != cookie.m_uids.end(); i++)
				cout << hex << *i << endl;
		}
		
		cout << "Terminating...." << endl;	
	}
	catch (exception &e)
	{
		cerr << e.what() << endl;
	}
		
	return 0;
}
